{
 "metadata": {
  "name": "",
  "signature": "sha256:b9ff21cab295a360df1bc3d77c2a46b73b900c673f617b1b983073d0f960bc28"
 },
 "nbformat": 3,
 "nbformat_minor": 0,
 "worksheets": [
  {
   "cells": [
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "%matplotlib inline\n",
      "import numpy as np\n",
      "import numpy.random as rnd\n",
      "import matplotlib.pyplot as plt, mpld3\n",
      "import operator\n",
      "plt.rcParams['figure.figsize'] = 12, 8\n",
      "mpld3.enable_notebook()"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [],
     "prompt_number": 1
    },
    {
     "cell_type": "heading",
     "level": 1,
     "metadata": {},
     "source": [
      "Quantum Computer"
     ]
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "def kron(*Xs):\n",
      "    Y = np.eye(1)\n",
      "    for X in Xs:\n",
      "        Y = np.kron(Y, X)\n",
      "    return Y\n",
      "\n",
      "def prod(iterable):\n",
      "    return reduce(operator.mul, iterable, 1)\n",
      "\n",
      "def randpsi(d):\n",
      "    psi = np.random.normal(size=d) + 1j * np.random.normal(size=d)\n",
      "    return psi / np.linalg.norm(psi)"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [],
     "prompt_number": 2
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "class QC(object):\n",
      "    def __init__(self, num_qubits):\n",
      "        \"\"\"starts out in state |0....0>\"\"\"\n",
      "        self.num_qubits = num_qubits\n",
      "        self.psi = np.zeros(2**num_qubits)\n",
      "        self.psi[0] = 1\n",
      "        \n",
      "    def _tensor(self, indices, Xs):\n",
      "        \"\"\"compute tensor product of operators Xs at indices\"\"\"\n",
      "        assert len(indices) == len(set(indices)), 'no repetitions'\n",
      "        assert all(i < self.num_qubits for i in indices), 'index out of bounds'\n",
      "        X = np.eye(1)\n",
      "        for i in range(self.num_qubits):\n",
      "            if i in indices:\n",
      "                X = np.kron(X, Xs[indices.index(i)])\n",
      "            else:\n",
      "                X = np.kron(X, np.eye(2))\n",
      "        return X\n",
      "    \n",
      "    def rdm(self, *indices):\n",
      "        \"\"\"reduced density matrix\"\"\"\n",
      "        indices = list(indices)\n",
      "        compl = [i for i in range(self.num_qubits) if i not in indices]\n",
      "        A = self.psi.reshape([2] * self.num_qubits).transpose(indices + compl).reshape([2**len(indices), 2**len(compl)])\n",
      "        return np.dot(A, A.T.conj())\n",
      "    \n",
      "    def H(self, idx):\n",
      "        \"\"\"Hadamard gate\"\"\"\n",
      "        H = [[1, 1], [1, -1]] / np.sqrt(2)\n",
      "        U = self._tensor([idx], [H])\n",
      "        self.psi = np.dot(U, self.psi)\n",
      "    \n",
      "    def NOT(self, idx):\n",
      "        \"\"\"NOT gate\"\"\"\n",
      "        NOT = [[0, 1], [1, 0]]\n",
      "        U = self._tensor([idx], [NOT])\n",
      "        self.psi = np.dot(U, self.psi)\n",
      "\n",
      "    def Z(self, idx):\n",
      "        \"\"\"Pauli Z-operator\"\"\"\n",
      "        Z = [[1, 0], [0, -1]]\n",
      "        U = self._tensor([idx], [Z])\n",
      "        self.psi = np.dot(U, self.psi)\n",
      "\n",
      "    def C(self, U, idx, control_idx):\n",
      "        \"\"\"controlled unitary\"\"\"\n",
      "        proj0 = [[1, 0], [0, 0]]\n",
      "        proj1 = [[0, 0], [0, 1]]\n",
      "        U = self._tensor([idx, control_idx], [np.eye(2), proj0]) + self._tensor([idx, control_idx], [U, proj1])\n",
      "        self.psi = np.dot(U, self.psi)        \n",
      "    \n",
      "    def CNOT(self, idx, control_idx):\n",
      "        \"\"\"controller NOT\"\"\"\n",
      "        NOT = [[0, 1], [1, 0]]\n",
      "        self.C(NOT, idx, control_idx)\n",
      "    \n",
      "    def measure(self, idx):\n",
      "        \"\"\"measure a qubit\"\"\"\n",
      "        proj0 = [[1, 0], [0, 0]]\n",
      "        psi0 = np.dot(self._tensor([idx], [proj0]), self.psi)\n",
      "        prob0 = np.linalg.norm(psi0)**2\n",
      "        if np.random.random() < prob0:\n",
      "            self.psi = psi0 / np.sqrt(prob0)\n",
      "            return 0\n",
      "        else:\n",
      "            self.psi = (self.psi - psi0) / np.sqrt(1 - prob0)\n",
      "            return 1"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [],
     "prompt_number": 3
    },
    {
     "cell_type": "heading",
     "level": 1,
     "metadata": {},
     "source": [
      "Teleportation"
     ]
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "KET_EPR = np.array([1, 0, 0, 1]) / np.sqrt(2)\n",
      "    \n",
      "def teleport(qc, ALICE_1, ALICE_2, BOB):\n",
      "    # teleportation protocol\n",
      "    assert np.allclose(qc.rdm(ALICE_2, BOB), np.outer(KET_EPR, KET_EPR))\n",
      "    qc.CNOT(ALICE_2, control_idx=ALICE_1)\n",
      "    qc.H(ALICE_1)\n",
      "    bits = [qc.measure(ALICE_1), qc.measure(ALICE_2)]   # ...these are sent to bob...\n",
      "    if bits[1]:\n",
      "        qc.NOT(BOB)\n",
      "    if bits[0]:\n",
      "        qc.Z(BOB)\n",
      "    return qc.rdm(BOB)\n",
      "\n",
      "# simple test\n",
      "qc = QC(num_qubits=3)\n",
      "\n",
      "# prepare initial state of Alice' first qubit\n",
      "qc.H(0)\n",
      "expected = qc.rdm(0)\n",
      "\n",
      "# prepare maximally entangled state between Alice' other qubit and Bob's qubit\n",
      "qc.H(1)\n",
      "qc.CNOT(2, control_idx=1)\n",
      "    \n",
      "got = teleport(qc, 0, 1, 2)\n",
      "assert np.allclose(expected, got)"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [],
     "prompt_number": 4
    },
    {
     "cell_type": "markdown",
     "metadata": {},
     "source": [
      "Teleport random states:"
     ]
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "%%time\n",
      "qc = QC(num_qubits=3)\n",
      "\n",
      "for _ in range(100):\n",
      "    # initialize QC with random state on first qubit and EPR pair between the other two\n",
      "    qc.psi = np.kron(randpsi(2), KET_EPR)\n",
      "    expected = qc.rdm(0)\n",
      "    \n",
      "    # teleport\n",
      "    teleport(qc, 0, 1, 2)\n",
      "    \n",
      "    # verify\n",
      "    got = qc.rdm(2)\n",
      "    assert np.allclose(expected, got)"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [
      {
       "output_type": "stream",
       "stream": "stdout",
       "text": [
        "Wall time: 119 ms\n"
       ]
      }
     ],
     "prompt_number": 5
    },
    {
     "cell_type": "heading",
     "level": 1,
     "metadata": {},
     "source": [
      "Entanglement Swapping"
     ]
    },
    {
     "cell_type": "markdown",
     "metadata": {},
     "source": [
      "Note that the \"entanglement\" (in fact, the entire reduced state) is preserved by the teleportation procedure."
     ]
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "%%time\n",
      "qc = QC(num_qubits=4)\n",
      "\n",
      "for _ in range(100):\n",
      "    # initialize QC with random state on first two qubits and EPR pair on the other two\n",
      "    qc.psi = np.kron(randpsi(4), KET_EPR)\n",
      "    expected = qc.rdm(0, 1)\n",
      "    \n",
      "    # teleport\n",
      "    teleport(qc, 1, 2, 3)\n",
      "    \n",
      "    # verify\n",
      "    got = qc.rdm(0, 3)\n",
      "    assert np.allclose(expected, got)"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [
      {
       "output_type": "stream",
       "stream": "stdout",
       "text": [
        "Wall time: 153 ms\n"
       ]
      }
     ],
     "prompt_number": 6
    },
    {
     "cell_type": "markdown",
     "metadata": {},
     "source": [
      "Thus given an EPR pair between Alice and Bob and another one between Bob and Charly we can get an EPR pair between Alice and Charly!"
     ]
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "qc = QC(num_qubits=4)\n",
      "ALICE, BOB_1, BOB_2, CHARLY = range(4)\n",
      "\n",
      "# initialize with two EPR pairs between Alice and Bob1 & between Bob2 and Charly, respectively\n",
      "qc.psi = np.kron(KET_EPR, KET_EPR)\n",
      "\n",
      "# teleport Bob1 via Bob2--Charly\n",
      "teleport(qc, BOB_1, BOB_2, CHARLY)\n",
      "\n",
      "# Alice--Charly are now entangled (and can do stuff like QKD, etc.)\n",
      "assert np.allclose(qc.rdm(0, 3), np.outer(KET_EPR, KET_EPR))"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [],
     "prompt_number": 7
    },
    {
     "cell_type": "heading",
     "level": 1,
     "metadata": {},
     "source": [
      "Deutsch-Josza Algorithm"
     ]
    },
    {
     "cell_type": "markdown",
     "metadata": {},
     "source": [
      "Truth tables of all constant and balanced functions of $n$ bits:"
     ]
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "from itertools import combinations\n",
      "\n",
      "def constant_fns(n):\n",
      "    yield [0] * (2**n)\n",
      "    yield [1] * (2**n)\n",
      "\n",
      "def balanced_fns(n):\n",
      "    inputs = range(2**n)\n",
      "    for s in combinations(inputs, 2**(n-1)):\n",
      "        yield [int(i in s) for i in inputs]\n",
      "\n",
      "def rand_balanced_fn(n):\n",
      "    inputs = range(2**n)\n",
      "    s = np.random.choice(inputs, 2**(n-1), replace=False)\n",
      "    return [int(i in s) for i in inputs]\n",
      "\n",
      "print list(constant_fns(2))\n",
      "print list(balanced_fns(2))\n",
      "print rand_balanced_fn(2)"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [
      {
       "output_type": "stream",
       "stream": "stdout",
       "text": [
        "[[0, 0, 0, 0], [1, 1, 1, 1]]\n",
        "[[1, 1, 0, 0], [1, 0, 1, 0], [1, 0, 0, 1], [0, 1, 1, 0], [0, 1, 0, 1], [0, 0, 1, 1]]\n",
        "[0, 1, 0, 1]\n"
       ]
      }
     ],
     "prompt_number": 8
    },
    {
     "cell_type": "markdown",
     "metadata": {},
     "source": [
      "Unitary oracle for a boolean function $f$:"
     ]
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "def U_oracle(n, f):\n",
      "    U = np.zeros(shape=(2**(n+1), 2**(n+1)))\n",
      "    for x in range(2**n):\n",
      "        for y in range(2):\n",
      "            z = (y + f[x]) % 2\n",
      "            U[x*2 + z, x*2 + y] = 1\n",
      "    return U\n",
      "\n",
      "U_oracle(1, [1, 0])"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [
      {
       "metadata": {},
       "output_type": "pyout",
       "prompt_number": 9,
       "text": [
        "array([[ 0.,  1.,  0.,  0.],\n",
        "       [ 1.,  0.,  0.,  0.],\n",
        "       [ 0.,  0.,  1.,  0.],\n",
        "       [ 0.,  0.,  0.,  1.]])"
       ]
      }
     ],
     "prompt_number": 9
    },
    {
     "cell_type": "markdown",
     "metadata": {},
     "source": [
      "Deutsch--Josza algorithm:"
     ]
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "def deutsch_josza(n, f):\n",
      "    # construct unitary oracle\n",
      "    U = U_oracle(n, f)\n",
      "    \n",
      "    # initialize in state |0>^{\\otimes n}  \\otimes  |1>\n",
      "    qc = QC(n+1)\n",
      "    qc.NOT(n)\n",
      "    \n",
      "    # hadamards on all qubits\n",
      "    for i in range(n+1):\n",
      "        qc.H(i)\n",
      "    \n",
      "    # oracle\n",
      "    qc.psi = np.dot(U, qc.psi)\n",
      "    \n",
      "    # hadamards on all qubits (except for last one, which does not matter anymore)\n",
      "    for i in range(n):\n",
      "        qc.H(i)\n",
      "    \n",
      "    # measure...\n",
      "    bits = [qc.measure(i) for i in range(n)]\n",
      "    is_const = all(b == 0 for b in bits)\n",
      "    return is_const\n",
      "\n",
      "assert deutsch_josza(2, [0, 0, 0, 0]) == True\n",
      "assert deutsch_josza(2, [0, 1, 0, 1]) == False"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [],
     "prompt_number": 10
    },
    {
     "cell_type": "markdown",
     "metadata": {},
     "source": [
      "Systematic Test:"
     ]
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "%%time\n",
      "n = 3\n",
      "for f in constant_fns(n):\n",
      "    assert deutsch_josza(n, f) == True\n",
      "for f in balanced_fns(n):\n",
      "    assert deutsch_josza(n, f) == False\n",
      "    \n",
      "n = 5\n",
      "for _ in range(100):\n",
      "    f = rand_balanced_fn(n)\n",
      "    assert deutsch_josza(n, f) == False"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [
      {
       "output_type": "stream",
       "stream": "stdout",
       "text": [
        "Wall time: 994 ms\n"
       ]
      }
     ],
     "prompt_number": 12
    }
   ],
   "metadata": {}
  }
 ]
}